"""Merge Methods

In the current implementation only fast-forward and a competent, but limited,
three-way merge algorithm are implemented. All user facing API calls should be
funneled through the :function:`select_merge_algorithm` function

.. note::

    In the current implementation, it is not possible to stop a merge in progress or
    to revert a bad merge commit. All revert like operations should be made by
    creating new branches from the last "good" state, after which new merge
    operations can be attempted (if desired.)
"""
from pathlib import Path

import lmdb

from .diff import WriterUserDiff, diff_envs, find_conflicts
from .records.commiting import (
    tmp_cmt_env,
    replace_staging_area_with_commit,
    replace_staging_area_with_refs,
    commit_records,
)
from .records.hashs import clear_stage_hash_records, backends_remove_in_process_data
from .records.heads import (
    get_staging_branch_head,
    get_branch_head_commit,
    set_staging_branch_head,
    set_branch_head_commit,
    release_writer_lock,
    acquire_writer_lock,
)


def select_merge_algorithm(message: str,
                           branchenv: lmdb.Environment,
                           stageenv: lmdb.Environment,
                           refenv: lmdb.Environment,
                           stagehashenv: lmdb.Environment,
                           master_branch: str,
                           dev_branch: str,
                           repo_path: Path,
                           *,
                           writer_uuid: str = 'MERGE_PROCESS') -> str:
    """Entry point to perform a merge.

    Automatically selects algorithm and does the operation if no conflicts are
    found. This call requires that the staging area status be "CLEAN", if
    a "DIRTY" staging environment is found, an RuntimeError will be thrown.

    Parameters
    ----------
    message : str
        user message describing the commit
    branchenv : lmdb.Environment
        where the branch references are stored
    stageenv : lmdb.Environment
        where the staging area is open
    refenv : lmdb.Environment
        where commit history is stored
    stagehashenv: lmdb.Environment
        where the stage hash environment data is stored
    master_branch : str
        name of the branch to serve as a merge master
    dev_branch : str
        name of the branch to use as the feature branch
    repo_path: Path
        path to the repository on disk
    writer_uuid : str, optional, kwarg only
        if the merge method is called from the repo level, the default writer
        lock `MERGE_PROCESS` is used to ensure that a writer is active. If
        called from within a write-enabled checkout, the writer lock is set to
        the writer_uuid of the writer checkout so that the lock can be acquired.

    Raises
    ------
    RuntimeError
        if the staging area is not `CLEAN` of other changes
    PermissionError
        if the writer lock is currently held
    ValueError
        if a conflict is found in a three way merge, no operation will be performed.

    Returns
    -------
    str
        commit hash of the merge if this was a successful operation.
    """
    current_head = get_staging_branch_head(branchenv)
    wDiffer = WriterUserDiff(stageenv=stageenv,
                             branchenv=branchenv,
                             refenv=refenv,
                             branch_name=current_head)
    if wDiffer.status() != 'CLEAN':
        e = RuntimeError(
            'Changes are currently pending in the staging area To avoid mangled '
            'histories, the staging area must exist in a clean state Please '
            'reset or commit any changes before the merge operation')
        raise e from None

    try:
        acquire_writer_lock(branchenv=branchenv, writer_uuid=writer_uuid)
    except PermissionError as e:
        raise e from None

    try:
        mHEAD = get_branch_head_commit(branchenv, branch_name=master_branch)
        dHEAD = get_branch_head_commit(branchenv, branch_name=dev_branch)
        branchHistory = wDiffer._determine_ancestors(mHEAD=mHEAD, dHEAD=dHEAD)

        if branchHistory.canFF is True:
            print('Selected Fast-Forward Merge Strategy')
            success = _fast_forward_merge(
                branchenv=branchenv,
                stageenv=stageenv,
                refenv=refenv,
                stagehashenv=stagehashenv,
                master_branch=master_branch,
                new_masterHEAD=branchHistory.devHEAD,
                repo_path=repo_path)
        else:
            print('Selected 3-Way Merge Strategy')
            success = _three_way_merge(
                message=message,
                master_branch=master_branch,
                masterHEAD=branchHistory.masterHEAD,
                dev_branch=dev_branch,
                devHEAD=branchHistory.devHEAD,
                ancestorHEAD=branchHistory.ancestorHEAD,
                branchenv=branchenv,
                stageenv=stageenv,
                refenv=refenv,
                stagehashenv=stagehashenv,
                repo_path=repo_path)

    except ValueError as e:
        raise e from None

    finally:
        if writer_uuid == 'MERGE_PROCESS':
            release_writer_lock(branchenv=branchenv, writer_uuid=writer_uuid)

    return success


# ------------------ Fast Forward Merge Methods -------------------------------


def _fast_forward_merge(branchenv: lmdb.Environment,
                        stageenv: lmdb.Environment,
                        refenv: lmdb.Environment,
                        stagehashenv: lmdb.Environment,
                        master_branch: str,
                        new_masterHEAD: str,
                        repo_path: Path) -> str:
    """Update branch head pointer to perform a fast-forward merge.

    This method does not check that it is safe to do this operation, all
    verification should happen before this point is reached

    Parameters
    ----------
    branchenv : lmdb.Environment
        db with the branch head pointers
    stageenv : lmdb.Environment
        db where the staging area records are stored.
    refenv : lmdb.Environment
        db where the merge commit records are stored.
    stagehashenv: lmdb.Environment
        db where the staged hash records are stored
    master_branch : str
        name of the merge_master branch which should be updated
    new_masterHEAD : str
        commit hash to update the master_branch name to point to.
    repo_path: Path
        path to the repository on disk.

    Returns
    -------
    str
        if successful, returns the commit hash the master branch name was
        updated to.
    """
    try:
        replace_staging_area_with_commit(
            refenv=refenv, stageenv=stageenv, commit_hash=new_masterHEAD)

        outBranchName = set_branch_head_commit(
            branchenv=branchenv, branch_name=master_branch, commit_hash=new_masterHEAD)
        set_staging_branch_head(branchenv=branchenv, branch_name=master_branch)

        backends_remove_in_process_data(repo_path=repo_path)
        clear_stage_hash_records(stagehashenv=stagehashenv)

    except ValueError as e:
        raise e from None

    return outBranchName


# ----------------------- Three-Way Merge Methods -----------------------------


def _three_way_merge(message: str,
                     master_branch: str,
                     masterHEAD: str,
                     dev_branch: str,
                     devHEAD: str,
                     ancestorHEAD: str,
                     branchenv: lmdb.Environment,
                     stageenv: lmdb.Environment,
                     refenv: lmdb.Environment,
                     stagehashenv: lmdb.Environment,
                     repo_path: Path) -> str:
    """Merge strategy with diff/patch computed from changes since last common ancestor.

    Parameters
    ----------
    message : str
        commit message to apply to this merge commit (specified by the user)
    master_branch : str
        name of the merge master branch
    masterHEAD : str
        commit hash of the merge master HEAD
    dev_branch : str
        name of the merge dev branch
    devHEAD : str
        commit hash of the merge dev HEAD
    ancestorHEAD : str
        commit hash of the nearest common ancestor which the merge_master and
        merge_dev branches both share in their commit history.
    branchenv : lmdb.Environment
        db where the branch head records are stored
    stageenv : lmdb.Environment
        db where the staging area records are stored.
    refenv : lmdb.Environment
        db where the merge commit records are stored.
    stagehashenv: lmdb.Environment
        db where the staged hash records are stored
    repo_path: Path
        path to the repository on disk.

    Returns
    -------
    str
        commit hash of the new merge commit if the operation was successful.

    Raises
    ------
    ValueError
        If a conflict is found, the operation will abort before completing.
    """
    with tmp_cmt_env(refenv, ancestorHEAD) as aEnv, tmp_cmt_env(
            refenv, masterHEAD) as mEnv, tmp_cmt_env(refenv, devHEAD) as dEnv:

        m_diff = diff_envs(aEnv, mEnv)
        d_diff = diff_envs(aEnv, dEnv)
        conflict = find_conflicts(m_diff, d_diff)
        if conflict.conflict is True:
            msg = f'HANGAR VALUE ERROR:: Merge ABORTED with conflict: {conflict}'
            raise ValueError(msg) from None

        with mEnv.begin(write=True) as txn:
            for k, _ in d_diff.deleted:
                txn.delete(k)
            for k, v in d_diff.mutated:
                txn.put(k, v, overwrite=True)
            for k, v in d_diff.added:
                txn.put(k, v, overwrite=True)

        dbcont = []
        with mEnv.begin(write=False) as txn:
            with txn.cursor() as cur:
                cur.first()
                for kv in cur.iternext(keys=True, values=True):
                    dbcont.append(kv)

    backends_remove_in_process_data(repo_path=repo_path)
    replace_staging_area_with_refs(stageenv=stageenv, sorted_content=dbcont)

    commit_hash = commit_records(
        message=message,
        branchenv=branchenv,
        stageenv=stageenv,
        refenv=refenv,
        repo_path=repo_path,
        is_merge_commit=True,
        merge_master=master_branch,
        merge_dev=dev_branch)

    clear_stage_hash_records(stagehashenv=stagehashenv)
    return commit_hash



import math
import struct
from typing import NamedTuple, List, Union, Tuple

import blosc
import numpy as np

from . import hangar_service_pb2
from ..utils import set_blosc_nthreads

set_blosc_nthreads()


def chunk_bytes(bytesData):
    """Slice a bytestring into subelements and store the data in a list

    Arguments
    ---------
        bytesData : bytes
            bytestring buffer of the array data

    Yields
    ------
    bytes
        data split into 32kb chunk sizes.
    """
    chunkSize = 32_000
    numIters = math.ceil(len(bytesData) / chunkSize)
    currentStart = 0
    currentEnd = chunkSize
    for i in range(numIters):
        yield bytesData[currentStart:currentEnd]
        currentStart += chunkSize
        currentEnd += chunkSize


def clientCommitChunkedIterator(commit: str, parentVal: bytes, specVal: bytes,
                                refVal: bytes) -> hangar_service_pb2.PushCommitRequest:
    """Generator splitting commit specs into chunks sent from client to server

    Parameters
    ----------
    commit : str
        commit hash which is being sent
    parentVal : bytes
        bytes representing the commits immediate parents
    specVal : bytes
        bytes representing the commit message/user specifications
    refVal : bytes
        bytes containing all records stored in the repository

    Yields
    ------
    hangar_service_pb2.PushCommitRequest
        Chunked generator of the PushCommitRequest protobuf.
    """
    commit_proto = hangar_service_pb2.CommitRecord(
        parent=parentVal,
        spec=specVal)
    byteSize = len(refVal)
    chunkIterator = chunk_bytes(refVal)
    for refChunk in chunkIterator:
        commit_proto.ref = refChunk
        request = hangar_service_pb2.PushCommitRequest(
            commit=commit,
            total_byte_size=byteSize,
            record=commit_proto)
        yield request


def tensorChunkedIterator(buf, uncomp_nbytes, pb2_request, *, err=None):

    compBytes = blosc.compress(
        buf, clevel=3, cname='blosclz', shuffle=blosc.NOSHUFFLE)

    request = pb2_request(
        comp_nbytes=len(compBytes),
        uncomp_nbytes=uncomp_nbytes,
        error=err)
    chunkIterator = chunk_bytes(compBytes)
    for dchunk in chunkIterator:
        request.raw_data = dchunk
        yield request


def missingHashIterator(commit, hash_bytes, err, pb2_func):
    comp_bytes = blosc.compress(
        hash_bytes, cname='zlib', clevel=3, typesize=1, shuffle=blosc.SHUFFLE)

    rpc_method = pb2_func(
        commit=commit,
        total_byte_size=len(comp_bytes),
        error=err)

    chunkIterator = chunk_bytes(comp_bytes)
    for bchunk in chunkIterator:
        rpc_method.hashs = bchunk
        yield rpc_method


def missingHashRequestIterator(commit, hash_bytes, pb2_func):
    comp_bytes = blosc.compress(
        hash_bytes, cname='zlib', clevel=3, typesize=1, shuffle=blosc.SHUFFLE)

    rpc_method = pb2_func(
        commit=commit,
        total_byte_size=len(comp_bytes))

    chunkIterator = chunk_bytes(comp_bytes)
    for bchunk in chunkIterator:
        rpc_method.hashs = bchunk
        yield rpc_method


# ------------------------ serialization formats -------------------------


class DataIdent(NamedTuple):
    digest: str
    schema: str


class DataRecord(NamedTuple):
    data: Union[np.ndarray, str, bytes]
    digest: str
    schema: str


def _serialize_arr(arr: np.ndarray) -> bytes:
    """
    dtype_num ndim dim1_size dim2_size ... dimN_size array_bytes
    """
    raw = struct.pack(
        f'<bb{len(arr.shape)}i{arr.nbytes}s',
        arr.dtype.num, arr.ndim, *arr.shape, arr.tobytes()
    )
    return raw


def _deserialize_arr(raw: bytes) -> np.ndarray:
    dtnum, ndim = struct.unpack('<bb', raw[0:2])
    end = 2 + (4 * ndim)
    arrshape = struct.unpack(f'<{ndim}i', raw[2:end])
    arr = np.frombuffer(raw, dtype=np.typeDict[dtnum], offset=end).reshape(arrshape)
    return arr


def _serialize_str(data: str) -> bytes:
    """
    data_bytes
    """
    return data.encode()


def _deserialize_str(raw: bytes) -> str:
    return raw.decode()


def _serialize_bytes(data: bytes) -> bytes:
    """
    data_bytes
    """
    return data


def _deserialize_bytes(data: bytes) -> bytes:
    return data


def serialize_ident(digest: str, schema: str) -> bytes:
    """
    len_digest len_schema digest_str schema_str
    """
    raw = struct.pack(
        f'<hh{len(digest)}s{len(schema)}s',
        len(digest), len(schema), digest.encode(), schema.encode()
    )
    return raw


def deserialize_ident(raw: bytes) -> DataIdent:
    digestLen, schemaLen = struct.unpack('<hh', raw[:4])
    rawdigest, rawschema = struct.unpack(f'<{digestLen}s{schemaLen}s', raw[4:])
    digest = rawdigest.decode()
    schema = rawschema.decode()
    return DataIdent(digest, schema)


def serialize_data(data: Union[np.ndarray, str, bytes]) -> Tuple[int, bytes]:
    if isinstance(data, np.ndarray):
        return (0, _serialize_arr(data))
    elif isinstance(data, str):
        return (2, _serialize_str(data))
    elif isinstance(data, bytes):
        return (3, _serialize_bytes(data))
    else:
        raise TypeError(type(data))


def deserialize_data(dtype_code: int, raw_data: bytes) -> Union[np.ndarray, str, bytes]:
    if dtype_code == 0:
        return _deserialize_arr(raw_data)
    elif dtype_code == 2:
        return _deserialize_str(raw_data)
    elif dtype_code == 3:
        return _deserialize_bytes(raw_data)
    else:
        raise ValueError(f'dtype_code unknown {dtype_code}')


def serialize_record(data: Union[np.ndarray, str, bytes], digest: str, schema: str) -> bytes:
    """
    dtype_code len_raw_ident len_raw_data raw_ident, raw_data
    """
    dtype_code, raw_data = serialize_data(data)
    raw_ident = serialize_ident(digest, schema)
    raw = struct.pack(
        f'<b2Q{len(raw_ident)}s{len(raw_data)}s',
        dtype_code, len(raw_ident), len(raw_data), raw_ident, raw_data
    )
    return raw


def deserialize_record(raw: bytes) -> DataRecord:
    identStart = 17  # 1 + 2 * 8 bytes
    dtype_code, identLen, dataLen = struct.unpack(f'<b2Q', raw[:identStart])
    identEnd = identStart + identLen
    arrEnd = identEnd + dataLen
    arr = deserialize_data(dtype_code, raw[identEnd:arrEnd])
    ident = deserialize_ident(raw[identStart:identEnd])
    return DataRecord(arr, ident.digest, ident.schema)


def serialize_record_pack(records: List[bytes]) -> bytes:
    """
    num_records len_rec1 raw_rec1 len_rec2 raw_rec2 ... len_recN raw_recN
    """
    raw_num_records = struct.pack(f'<i', len(records))
    raw_records = [b''.join([struct.pack(f'<Q', len(rec)), rec]) for rec in records]
    return b''.join([raw_num_records, *raw_records])


def deserialize_record_pack(raw: bytes) -> List[bytes]:
    numRecords = struct.unpack(f'<i', raw[:4])[0]
    cursorPos, recs = 4, []
    for i in range(numRecords):
        lenRec = struct.unpack(f'<Q', raw[cursorPos:cursorPos+8])[0]
        recs.append(raw[cursorPos+8:cursorPos+8+lenRec])
        cursorPos += (8 + lenRec)
    return recs



import configparser
import os
from pathlib import Path
import platform
import shutil
import tempfile
import warnings
from typing import MutableMapping, Optional

import lmdb

from . import __version__
from .constants import (
    CONFIG_USER_NAME,
    DIR_DATA_REMOTE,
    DIR_DATA_STAGE,
    DIR_DATA_STORE,
    DIR_DATA,
    LMDB_BRANCH_NAME,
    LMDB_HASH_NAME,
    LMDB_REF_NAME,
    LMDB_SETTINGS,
    LMDB_STAGE_HASH_NAME,
    LMDB_STAGE_REF_NAME,
    README_FILE_NAME,
)
from .records.commiting import unpack_commit_ref
from .records.heads import (
    create_branch,
    get_branch_head_commit,
    get_staging_branch_head,
    set_staging_branch_head,
)
from .records.parsing import repo_version_raw_spec_from_raw_string
from .records.vcompat import (
    is_repo_software_version_compatible,
    set_repository_software_version,
    startup_check_repo_version,
)
from .utils import readme_contents, is_64bits


class Environments(object):

    def __init__(self, pth: Path):

        self.repo_path: Path = pth
        self.refenv: Optional[lmdb.Environment] = None
        self.hashenv: Optional[lmdb.Environment] = None
        self.stageenv: Optional[lmdb.Environment] = None
        self.branchenv: Optional[lmdb.Environment] = None
        self.stagehashenv: Optional[lmdb.Environment] = None
        self.cmtenv: MutableMapping[str, lmdb.Environment] = {}
        self._startup()

    @property
    def repo_is_initialized(self) -> bool:
        """Property to check if the repository is initialized, read-only attribute

        Returns
        -------
        bool
            True if repo environments are initialized, False otherwise
        """
        ret = True if isinstance(self.refenv, lmdb.Environment) else False
        return ret

    def _startup(self) -> bool:
        """When first access to the Repo starts, attempt to open the db envs.

        This function is designed to fail if a repository does not exist at the
        :py:attribute:`repo_path` which is specified, so the user can
        explicitly choose to initialize the repo. Once opened, the lmdb
        environments should not be closed until the program terminates.

        Returns
        -------
        bool False if no repository exists at the given path, otherwise True

        Warns
        -----
        UserWarning Should the repository not exist at the provided repo path.

        Raises
        ------
        RuntimeError If the repository version is not compatible with the
        current software.
        """

        if not self.repo_path.joinpath(LMDB_BRANCH_NAME).is_file():
            msg = f'No repository exists at {self.repo_path}, please use `repo.init()` method'
            warnings.warn(msg, UserWarning)
            return False

        if not is_64bits():
            raise OSError(f'Hangar cannot run on 32 bit machines')

        repo_ver = startup_check_repo_version(self.repo_path)
        curr_ver = repo_version_raw_spec_from_raw_string(v_str=__version__)
        if not is_repo_software_version_compatible(repo_ver, curr_ver):
            msg = f'repository written version: {repo_ver} is not comatible '\
                  f'with the current Hangar software version: {curr_ver}'
            raise RuntimeError(msg)

        self._open_environments()
        return True

    def init_repo(self,
                  user_name: str,
                  user_email: str,
                  remove_old: bool = False) -> Path:
        """Create a new hangar repositiory at the specified environment path.

        Parameters
        ----------
        user_name : str
            Name of the repository user.
        user_email : str
            Email address of the respository user.
        remove_old : bool, optional(default value = False)
            DEVELOPER USE ONLY --- Remove all data and records stored in the
            repository if this opetion is enabled, defaults to False.

        Returns
        -------
        Path
            The path to the newly created repository on disk.

        Raises
        ------
        OSError
            If a hangar repository exists at the specified path, and `remove_old`
            was not set to ``True``.
        """
        if self.repo_path.joinpath(LMDB_BRANCH_NAME).is_file():
            if remove_old is True:
                shutil.rmtree(str(self.repo_path))
            else:
                raise OSError(f'Hangar Directory: {self.repo_path} already exists')

        self.repo_path.mkdir()
        self.repo_path.joinpath(DIR_DATA_STORE).mkdir()
        self.repo_path.joinpath(DIR_DATA_STAGE).mkdir()
        self.repo_path.joinpath(DIR_DATA_REMOTE).mkdir()
        self.repo_path.joinpath(DIR_DATA).mkdir()
        print(f'Hangar Repo initialized at: {self.repo_path}')

        userConf = {'USER': {'name': user_name, 'email': user_email}}
        CFG = configparser.ConfigParser()
        CFG.read_dict(userConf)
        with self.repo_path.joinpath(CONFIG_USER_NAME).open('w') as f:
            CFG.write(f)

        readmeTxt = readme_contents(user_name, user_email)
        with self.repo_path.joinpath(README_FILE_NAME).open('w') as f:
            f.write(readmeTxt.getvalue())

        self._open_environments()
        set_repository_software_version(branchenv=self.branchenv, ver_str=__version__)
        create_branch(self.branchenv, 'master', '')
        set_staging_branch_head(self.branchenv, 'master')
        return self.repo_path

    def checkout_commit(self, branch_name: str = '', commit: str = '') -> str:
        """Set up db environment with unpacked commit ref records.

        Parameters
        ----------
        branch_name : str, optional
            name of the branch to read, defaults to ''
        commit : str, optional
            name of the commit to read, defaults to ''

        Returns
        -------
        str
            commit hash which was checked out
        """
        if commit != '':
            commit_hash = commit
            txt = f' * Checking out COMMIT: {commit_hash}'
        elif branch_name != '':
            commit_hash = get_branch_head_commit(self.branchenv, branch_name)
            txt = f' * Checking out BRANCH: {branch_name} with current HEAD: {commit_hash}'
        else:
            head_branch = get_staging_branch_head(self.branchenv)
            commit_hash = get_branch_head_commit(self.branchenv, head_branch)
            txt = f'\n Neither BRANCH or COMMIT specified.'\
                  f'\n * Checking out writing HEAD BRANCH: {head_branch}'
        print(txt)

        # On UNIX-like system, an open process still retains ability to
        # interact with disk space allocated to a file when it is removed from
        # disk. Windows does not, and will not allow file to be removed if a
        # process is interacting with it. While the CM form is cleaner, this
        # hack allows similar usage on Windows platforms.

        if platform.system() != 'Windows':
            with tempfile.TemporaryDirectory() as tempD:
                tmpDF = os.path.join(tempD, f'{commit_hash}.lmdb')
                tmpDB = lmdb.open(path=tmpDF, **LMDB_SETTINGS)
                unpack_commit_ref(self.refenv, tmpDB, commit_hash)
                self.cmtenv[commit_hash] = tmpDB
        else:
            tempD = tempfile.mkdtemp()
            tmpDF = os.path.join(tempD, f'{commit_hash}.lmdb')
            tmpDB = lmdb.open(path=tmpDF, **LMDB_SETTINGS)
            unpack_commit_ref(self.refenv, tmpDB, commit_hash)
            self.cmtenv[commit_hash] = tmpDB

        return commit_hash

    def _open_environments(self):
        """Open the standard lmdb databases at the repo path.

        If any commits are checked out (in an unpacked state), read those in as
        well.
        """
        ref_pth = str(self.repo_path.joinpath(LMDB_REF_NAME))
        hash_pth = str(self.repo_path.joinpath(LMDB_HASH_NAME))
        stage_pth = str(self.repo_path.joinpath(LMDB_STAGE_REF_NAME))
        branch_pth = str(self.repo_path.joinpath(LMDB_BRANCH_NAME))
        stagehash_pth = str(self.repo_path.joinpath(LMDB_STAGE_HASH_NAME))

        self.refenv = lmdb.open(path=ref_pth, **LMDB_SETTINGS)
        self.hashenv = lmdb.open(path=hash_pth, **LMDB_SETTINGS)
        self.stageenv = lmdb.open(path=stage_pth, **LMDB_SETTINGS)
        self.branchenv = lmdb.open(path=branch_pth, **LMDB_SETTINGS)
        self.stagehashenv = lmdb.open(path=stagehash_pth, **LMDB_SETTINGS)

    def _close_environments(self):

        self.refenv.close()
        self.hashenv.close()
        self.stageenv.close()
        self.branchenv.close()
        self.stagehashenv.close()
        for env in self.cmtenv.values():
            if platform.system() == 'Windows':
                envpth = env.path()
                env.close()
                os.remove(envpth)
            else:
                env.close()



from .hashmachine import hash_func_from_tcode
from .column_parsers import *
from .recordstructs import (
    CompatibleData,
    ColumnSchemaKey,
    FlatColumnDataKey,
    NestedColumnDataKey,
    DataRecordVal,
)


__all__ = column_parsers.__all__ + [
    'hash_func_from_tcode',
    'CompatibleData',
    'ColumnSchemaKey',
    'FlatColumnDataKey',
    'NestedColumnDataKey',
    'DataRecordVal',
]



from .descriptors import OneOf, String, checkedmeta
from ..records import hash_func_from_tcode


@OneOf(['flat', 'nested'])
class ColumnLayout(String):
    pass


@OneOf(['str', 'ndarray', 'bytes'])
class ColumnDType(String):
    pass


@OneOf(['1'])
class SchemaHasherTcode(String):
    pass


class ColumnBase(metaclass=checkedmeta):
    _column_layout = ColumnLayout()
    _column_type = ColumnDType()
    _schema_hasher_tcode = SchemaHasherTcode()

    def __init__(
            self,
            column_layout,
            column_type,
            data_hasher_tcode,
            schema_hasher_tcode=None,
            *args, **kwargs
    ):
        if schema_hasher_tcode is None:
            schema_hasher_tcode = '1'

        self._column_layout = column_layout
        self._column_type = column_type
        self._schema_hasher_tcode = schema_hasher_tcode
        self._data_hasher_tcode = data_hasher_tcode
        self._schema_attributes = [
            '_column_layout',
            '_column_type',
            '_schema_hasher_tcode',
            '_data_hasher_tcode',
        ]
        self._schema_hasher_func = hash_func_from_tcode(self._schema_hasher_tcode)
        self._data_hasher_func = hash_func_from_tcode(self._data_hasher_tcode)
        self._hidden_be_opts = None

    @property
    def _beopts(self):
        from ..backends import BACKEND_OPTIONS_MAP
        if self._hidden_be_opts is None:
            self._hidden_be_opts = BACKEND_OPTIONS_MAP[self.backend](
                backend_options=self.backend_options,
                dtype=self.dtype,
                shape=(self.shape if hasattr(self, '_shape') else None))
        return self._hidden_be_opts

    @_beopts.deleter
    def _beopts(self):
        self._hidden_be_opts = None

    @_beopts.setter
    def _beopts(self, backend_options):
        from ..backends import BACKEND_OPTIONS_MAP
        self._hidden_be_opts = BACKEND_OPTIONS_MAP[self.backend](
            backend_options=backend_options,
            dtype=self.dtype,
            shape=(self.shape if hasattr(self, '_shape') else None))

    @property
    def column_layout(self):
        return self._column_layout

    @property
    def column_type(self):
        return self._column_type

    @property
    def schema_hasher_tcode(self):
        return self._schema_hasher_tcode

    @property
    def schema(self):
        schema_dict = {}
        public_attr_names = [attr.lstrip('_') for attr in self._schema_attributes]
        for attr in public_attr_names:
            schema_dict[attr] = getattr(self, f'_{attr}')
        return schema_dict

    def schema_hash_digest(self):
        return self._schema_hasher_func(self.schema)

    def backend_from_heuristics(self, *args, **kwargs):
        raise NotImplementedError

    def verify_data_compatible(self, *args, **kwargs):
        raise NotImplementedError

    @property
    def data_hasher_tcode(self):
        return self._data_hasher_tcode

    def data_hash_digest(self, *args, **kwargs):
        raise NotImplementedError




"""Constructor and Interaction Class for Columns
"""
from contextlib import ExitStack
from pathlib import Path
from typing import Iterable, List, Mapping, Optional, Tuple, Union, Dict

import lmdb

from .common import ColumnTxn
from .constructors import (
    generate_flat_column, generate_nested_column, column_type_object_from_schema
)
from ..records import (
    schema_db_key_from_column,
    schema_hash_db_key_from_digest,
    schema_column_record_from_db_key,
    schema_spec_from_db_val,
    dynamic_layout_data_record_db_start_range_key,
)
from ..records.queries import RecordQuery
from ..op_state import writer_checkout_only
from ..txnctx import TxnRegister

KeyType = Union[str, int]
ModifierTypes = Union['NestedSampleReader', 'FlatSubsampleReader']


class Columns:
    """Common access patterns and initialization/removal of columns in a checkout.

    This object is the entry point to all data stored in their
    individual columns. Each column. contains a common schema which dictates
    the general shape, dtype, and access patters which the backends optimize
    access for. The methods contained within allow us to create, remove, query,
    and access these collections of common data pieces.
    """

    def __init__(self,
                 mode: str,
                 repo_pth: Path,
                 columns: Dict[str, ModifierTypes],
                 hashenv: Optional[lmdb.Environment] = None,
                 dataenv: Optional[lmdb.Environment] = None,
                 stagehashenv: Optional[lmdb.Environment] = None,
                 txnctx: Optional[ColumnTxn] = None):
        """Developer documentation for init method.

        .. warning::

            This class should not be instantiated directly. Instead use the factory
            functions :py:meth:`_from_commit` or :py:meth:`_from_staging` to return
            a pre-initialized class instance appropriately constructed for either a
            read-only or write-enabled checkout.

        Parameters
        ----------
        mode : str
            one of 'r' or 'a' to indicate read or write mode
        repo_pth : Path
            path to the repository on disk
        columns : Mapping[str, Union[ArraysetDataReader, ArraysetDataWriter]]
            dictionary of ArraysetData objects
        hashenv : Optional[lmdb.Environment]
            environment handle for hash records
        dataenv : Optional[lmdb.Environment]
            environment handle for the unpacked records. `data` is means to refer to
            the fact that the stageenv is passed in for for write-enabled, and a
            cmtrefenv for read-only checkouts.
        stagehashenv : Optional[lmdb.Environment]
            environment handle for newly added staged data hash records.
        txnctx: Optional[ColumnTxn]
            class implementing context managers to handle lmdb transactions
        """
        self._stack: Optional[ExitStack] = None
        self._is_conman_counter = 0
        self._mode = mode
        self._repo_pth = repo_pth
        self._columns = columns

        self._hashenv = hashenv
        self._dataenv = dataenv
        self._stagehashenv = stagehashenv
        self._txnctx = txnctx

    def _open(self):
        for v in self._columns.values():
            v._open()

    def _close(self):
        for v in self._columns.values():
            v._close()

    def _destruct(self):
        if isinstance(self._stack, ExitStack):
            self._stack.close()
        self._close()
        for column in self._columns.values():
            column._destruct()
        for attr in list(self.__dict__.keys()):
            delattr(self, attr)

    def __getattr__(self, name):
        """Raise permission error after checkout is closed.

         Only runs after a call to :meth:`_destruct`, which is responsible
         for deleting all attributes from the object instance.
        """
        try:
            self.__getattribute__('_mode')  # once checkout is closed, this won't exist.
        except AttributeError:
            err = (f'Unable to operate on past checkout objects which have been '
                   f'closed. No operation occurred. Please use a new checkout.')
            raise PermissionError(err) from None
        return self.__getattribute__(name)

# ------------- Methods Available To Both Read & Write Checkouts ------------------

    def _repr_pretty_(self, p, cycle):
        res = f'Hangar {self.__class__.__qualname__}\
                \n    Writeable         : {False if self._mode == "r" else True}\
                \n    Number of Columns : {len(self)}\
                \n    Column Names / Partial Remote References:\
                \n      - ' + '\n      - '.join(
            f'{asetn} / {aset.contains_remote_references}'
            for asetn, aset in self._columns.items())
        p.text(res)

    def __repr__(self):
        res = f'{self.__class__}('\
              f'repo_pth={self._repo_pth}, '\
              f'columns={self._columns}, '\
              f'mode={self._mode})'
        return res

    def _ipython_key_completions_(self):
        """Let ipython know that any key based access can use the column keys

        Since we don't want to inherit from dict, nor mess with `__dir__` for
        the sanity of developers, this is the best way to ensure users can
        autocomplete keys.

        Returns
        -------
        list
            list of strings, each being one of the column keys for access.
        """
        return self.keys()

    def __getitem__(self, key: str) -> ModifierTypes:
        """Dict style access to return the column object with specified key/name.

        Parameters
        ----------
        key : string
            name of the column object to get.

        Returns
        -------
        ModifierTypes
            The object which is returned depends on the mode of checkout
            specified. If the column was checked out with write-enabled,
            return writer object, otherwise return read only object.
        """
        try:
            return self._columns[key]
        except KeyError:
            raise KeyError(f'No column exists with name: {key}')

    def __contains__(self, key: str) -> bool:
        """Determine if a column with a particular name is stored in the checkout

        Parameters
        ----------
        key : str
            name of the column to check for

        Returns
        -------
        bool
            True if a column with the provided name exists in the checkout,
            otherwise False.
        """
        return True if key in self._columns else False

    def __len__(self) -> int:
        """Get the number of column columns contained in the checkout.
        """
        return len(self._columns)

    def __iter__(self) -> Iterable[str]:
        return iter(self._columns)

    @property
    def _is_conman(self):
        return bool(self._is_conman_counter)

    def _any_is_conman(self) -> bool:
        """Determine if self or any contains column class is conman.

        Returns
        -------
        bool
            [description]
        """
        res = any([self._is_conman, *[x._is_conman for x in self._columns.values()]])
        return res

    def __enter__(self):
        with ExitStack() as stack:
            for asetN in list(self._columns.keys()):
                stack.enter_context(self._columns[asetN])
            self._is_conman_counter += 1
            self._stack = stack.pop_all()
        return self

    def __exit__(self, *exc):
        self._is_conman_counter -= 1
        self._stack.close()

    @property
    def iswriteable(self) -> bool:
        """Bool indicating if this column object is write-enabled. Read-only attribute.
        """
        return False if self._mode == 'r' else True

    @property
    def contains_remote_references(self) -> Mapping[str, bool]:
        """Dict of bool indicating data reference locality in each column.

        Returns
        -------
        Mapping[str, bool]
            For each column name key, boolean value where False indicates all
            samples in column exist locally, True if some reference remote
            sources.
        """
        res = {}
        for asetn, aset in self._columns.items():
            res[asetn] = aset.contains_remote_references
        return res

    @property
    def remote_sample_keys(self) -> Mapping[str, Iterable[Union[int, str]]]:
        """Determine columns samples names which reference remote sources.

        Returns
        -------
        Mapping[str, Iterable[Union[int, str]]]
            dict where keys are column names and values are iterables of
            samples in the column containing remote references
        """
        res = {}
        for asetn, aset in self._columns.items():
            res[asetn] = aset.remote_reference_keys
        return res

    def keys(self) -> List[str]:
        """list all column keys (names) in the checkout

        Returns
        -------
        List[str]
            list of column names
        """
        return list(self._columns.keys())

    def values(self) -> Iterable[ModifierTypes]:
        """yield all column object instances in the checkout.

        Yields
        -------
        Iterable[ModifierTypes]
            Generator of ColumnData accessor objects (set to read or write mode
            as appropriate)
        """
        for asetN in list(self._columns.keys()):
            asetObj = self._columns[asetN]
            yield asetObj

    def items(self) -> Iterable[Tuple[str, ModifierTypes]]:
        """generator providing access to column_name, :class:`Columns`

        Yields
        ------
        Iterable[Tuple[str, ModifierTypes]]
            returns two tuple of all all column names/object pairs in the checkout.
        """
        for asetN in list(self._columns.keys()):
            asetObj = self._columns[asetN]
            yield (asetN, asetObj)

    def get(self, name: str) -> ModifierTypes:
        """Returns a column access object.

        This can be used in lieu of the dictionary style access.

        Parameters
        ----------
        name : str
            name of the column to return

        Returns
        -------
        ModifierTypes
            ColumnData accessor (set to read or write mode as appropriate) which
            governs interaction with the data
        """
        return self[name]

    # -------------------- Writer-Enabled Methods Only ------------------------

    @writer_checkout_only
    def __delitem__(self, key: str) -> str:
        """remove a column and all data records if write-enabled process.

        Parameters
        ----------
        key : str
            Name of the column to remove from the repository. This will remove
            all records from the staging area (though the actual data and all
            records are still accessible) if they were previously committed

        Returns
        -------
        str
            If successful, the name of the removed column.

        Raises
        ------
        PermissionError
            If any enclosed column is opened in a connection manager.
        """
        if self._any_is_conman():
            raise PermissionError(
                'Not allowed while any columns class is opened in a context manager')
        return self.delete(key)

    @writer_checkout_only
    def delete(self, column: str) -> str:
        """remove the column and all data contained within it.

        Parameters
        ----------
        column : str
            name of the column to remove

        Returns
        -------
        str
            name of the removed column

        Raises
        ------
        PermissionError
            If any enclosed column is opened in a connection manager.
        KeyError
            If a column does not exist with the provided name
        """
        if self._any_is_conman():
            raise PermissionError(
                'Not allowed while any columns class is opened in a context manager')

        with ExitStack() as stack:
            datatxn = TxnRegister().begin_writer_txn(self._dataenv)
            stack.callback(TxnRegister().commit_writer_txn, self._dataenv)

            if column not in self._columns:
                e = KeyError(f'Cannot remove: {column}. Key does not exist.')
                raise e from None

            column_layout = self._columns[column].column_layout
            columnSchemaKey = schema_db_key_from_column(column, layout=column_layout)
            column_record = schema_column_record_from_db_key(columnSchemaKey)
            startRangeKey = dynamic_layout_data_record_db_start_range_key(column_record)

            self._columns[column]._close()
            self._columns.__delitem__(column)
            with datatxn.cursor() as cursor:
                cursor.first()
                recordsExist = cursor.set_range(startRangeKey)
                while recordsExist:
                    k = cursor.key()
                    if k.startswith(startRangeKey):
                        recordsExist = cursor.delete()
                    else:
                        recordsExist = False
            datatxn.delete(columnSchemaKey)

        return column

    @classmethod
    def _from_staging_area(cls, repo_pth, hashenv, stageenv, stagehashenv):
        """INTERNAL USE ONLY

        Class method factory to checkout :class:`Columns` in write mode

        Once you get here, we assume the write lock verification has
        passed, and that write operations are safe to perform.

        Parameters
        ----------
        repo_pth : Path
            directory path to the hangar repository on disk
        hashenv : lmdb.Environment
            environment where tensor data hash records are open in write mode.
        stageenv : lmdb.Environment
            environment where staging records (dataenv) are opened in write mode.
        stagehashenv: lmdb.Environment
            environment where the staged hash records are stored in write mode

        Returns
        -------
        :class:`~column.Columns`
            Interface class with write-enabled attributes activate which contains
            live column data accessors in `write` mode.
        """
        columns = {}
        txnctx = ColumnTxn(stageenv, hashenv, stagehashenv)
        query = RecordQuery(stageenv)
        stagedSchemaSpecs = query.schema_specs()

        staged_col_schemas = {}
        with txnctx.read() as r_txn:
            # need to do some conversions here...
            # ref record digest -> hash db key -> schema spec dict -> schema obj
            for column_record, schema_digest_rec in stagedSchemaSpecs.items():
                hashSchemaKey = schema_hash_db_key_from_digest(schema_digest_rec.digest)
                hashSchemaVal = r_txn.hashTxn.get(hashSchemaKey)
                schema_dict = schema_spec_from_db_val(hashSchemaVal)
                schema = column_type_object_from_schema(schema_dict)
                staged_col_schemas[column_record] = schema

        for column_record, schema in staged_col_schemas.items():
            if column_record.layout == 'nested':
                column = generate_nested_column(
                    txnctx=txnctx, column_name=column_record.column,
                    path=repo_pth, schema=schema, mode='a')
            else:
                column = generate_flat_column(
                    txnctx=txnctx, column_name=column_record.column,
                    path=repo_pth, schema=schema, mode='a')
            columns[column_record.column] = column

        return cls(mode='a',
                   repo_pth=repo_pth,
                   columns=columns,
                   hashenv=hashenv,
                   dataenv=stageenv,
                   stagehashenv=stagehashenv,
                   txnctx=txnctx)

    @classmethod
    def _from_commit(cls, repo_pth, hashenv, cmtrefenv):
        """INTERNAL USE ONLY

        Class method factory to checkout :class:`.Columns` in read-only mode

        For read mode, no locks need to be verified, but construction should
        occur through this interface only.

        Parameters
        ----------
        repo_pth : Path
            directory path to the hangar repository on disk
        hashenv : lmdb.Environment
            environment where tensor data hash records are open in read-only mode.
        cmtrefenv : lmdb.Environment
            environment where staging checkout records are opened in read-only mode.

        Returns
        -------
        :class:`~column.Columns`
            Interface class with write-enabled attributes deactivated which
            contains live column data accessors in `read-only` mode.
        """
        columns = {}
        txnctx = ColumnTxn(cmtrefenv, hashenv, None)
        query = RecordQuery(cmtrefenv)
        cmtSchemaSpecs = query.schema_specs()

        cmt_col_schemas = {}
        with txnctx.read() as r_txn:
            # need to do some conversions here...
            # ref record digest -> hash db key -> schema spec dict -> schema obj
            for column_record, schema_digest_rec in cmtSchemaSpecs.items():
                hashSchemaKey = schema_hash_db_key_from_digest(schema_digest_rec.digest)
                hashSchemaVal = r_txn.hashTxn.get(hashSchemaKey)
                schema_dict = schema_spec_from_db_val(hashSchemaVal)
                schema = column_type_object_from_schema(schema_dict)
                cmt_col_schemas[column_record] = schema

        for column_record, schema in cmt_col_schemas.items():
            if column_record.layout == 'nested':
                column = generate_nested_column(
                    txnctx=txnctx, column_name=column_record.column,
                    path=repo_pth, schema=schema, mode='r')
            else:
                column = generate_flat_column(
                    txnctx=txnctx, column_name=column_record.column,
                    path=repo_pth, schema=schema, mode='r')
            columns[column_record.column] = column

        return cls(mode='r',
                   repo_pth=repo_pth,
                   columns=columns,
                   hashenv=None,
                   dataenv=None,
                   stagehashenv=None,
                   txnctx=None)



from contextlib import contextmanager
from typing import Optional

import lmdb

from ..txnctx import TxnRegister


class ColumnTxn(object):
    """Provides context manager ready methods to handle lmdb transactions.

    In order to prevent passing around lmdb.Environment objects, we instantiate
    this class once for each column column and pass weakref proxy handels
    around to reference this object. Calling open / close methods (or using the
    ``with`` style methods) initializes transactions for the appropraite
    environments which are stored in instance attributes for access by the
    caller.
    """

    __slots__ = ('stagehashenv', 'dataenv', 'hashenv', 'hashTxn',
                 'dataTxn', 'stageHashTxn', '_TxnRegister', '__weakref__')

    def __init__(self, dataenv, hashenv, stagehashenv):

        self._TxnRegister = TxnRegister()
        self.stagehashenv = stagehashenv
        self.dataenv = dataenv
        self.hashenv = hashenv

        self.hashTxn: Optional[lmdb.Transaction] = None
        self.dataTxn: Optional[lmdb.Transaction] = None
        self.stageHashTxn: Optional[lmdb.Transaction] = None

    @property
    def _debug_(self):  # pragma: no cover
        return {
            f'__class__': self.__class__,
            f'_TxnRegister': self._TxnRegister._debug_,
            f'dataenv': self.dataenv,
            f'hashenv': self.hashenv,
            f'hashTxn': self.hashTxn,
            f'dataTxn': self.dataTxn,
            f'stageHashTxn': self.stageHashTxn,
        }

    def open_read(self):
        """Manually open read-only transactions, caller responsible for closing.
        """
        self.hashTxn = self._TxnRegister.begin_reader_txn(self.hashenv)
        self.dataTxn = self._TxnRegister.begin_reader_txn(self.dataenv)
        return self

    def close_read(self):
        """Manually close read-only transactions, must be called after manual open.
        """
        self.hashTxn = self._TxnRegister.abort_reader_txn(self.hashenv)
        self.dataTxn = self._TxnRegister.abort_reader_txn(self.dataenv)

    def open_write(self):
        """Manually open write-enabled transactions, caller responsible for closing.
        """
        self.hashTxn = self._TxnRegister.begin_writer_txn(self.hashenv)
        self.dataTxn = self._TxnRegister.begin_writer_txn(self.dataenv)
        self.stageHashTxn = self._TxnRegister.begin_writer_txn(self.stagehashenv)
        return self

    def close_write(self):
        """Manually close write-enabled transactions, must be called after manual open.
        """
        self.hashTxn = self._TxnRegister.commit_writer_txn(self.hashenv)
        self.dataTxn = self._TxnRegister.commit_writer_txn(self.dataenv)
        self.stageHashTxn = self._TxnRegister.commit_writer_txn(self.stagehashenv)

    @contextmanager
    def read(self):
        """Use ``with`` style context manager to open read-only transaction.

        Transaction is automatically closed for the caller irregardless of any
        application exceptions.
        """
        try:
            yield self.open_read()
        finally:
            self.close_read()

    @contextmanager
    def write(self):
        """Use ``with`` style context manager to open write-enabled transaction.

        Transaction is automatically closed for the caller irregardless of any
        application exceptions.
        """
        try:
            yield self.open_write()
        finally:
            self.close_write()


def open_file_handles(backends, path, mode, schema, *, remote_operation=False):
    """Open backend accessor file handles for reading

    Parameters
    ----------
    backends : Set[str]
        if ``mode == 'r'`` then this should be the used backend format
        codes in the column. if ``mode == 'a'``, then this should be a
        list of the allowed backend format codes this schema can feasably
        write to.
    path : Path
        path to the hangar repository on disk
    mode : str
        one of ['r', 'a'] indicating read or write mode to open backends in.
    schema : ColumnDefinitionTypes
        schema spec so required values can be filled in to backend openers.

    Returns
    -------
    AccessorMapType
        dict mapping backend format codes to initialized instances of each
        read-only backend.
    """
    from ..backends import BACKEND_ACCESSOR_MAP

    fhandles = {}
    for be, accessor in BACKEND_ACCESSOR_MAP.items():
        if be in backends:
            if accessor is None:
                continue

            init_requires = schema._beopts.init_requires
            # TODO rework names for this hack
            kwargs = {}
            for arg in init_requires:
                if arg == 'repo_path':
                    kwargs[arg] = path
                elif arg == 'schema_shape':
                    kwargs[arg] = schema.shape
                elif arg == 'schema_dtype':
                    kwargs[arg] = schema.dtype

            fhandles[be] = accessor(**kwargs)
            fhandles[be].open(mode=mode, remote_operation=remote_operation)

    if mode == 'a':
        if schema.backend in fhandles:
            fhandles[schema.backend].backend_opts = schema.backend_options
    return fhandles



"""
Portions of this code have been taken and modified from the "PyTables" project.

URL:      https://github.com/PyTables/PyTables
File:     tables/leaf.py
Commit:   1e7b14e87507c2392265321fe18b2f1f5920ea7f
Accessed: 23 JAN 2020

PyTables License
-------------------------------------------------------------------------------
License: BSD
URL:     https://github.com/PyTables/PyTables/blob/1e7b14e875/LICENSE.txt
"""
import numpy as np
import math


SizeType = np.int64


def _csformula(expected_mb):
    """Return the fitted chunksize for expected_mb."""

    # For a basesize of 8 KB, this will return:
    # 8 KB for datasets <= 1 MB
    # 1 MB for datasets >= 10 TB
    basesize = 8 * 1024  # 8 KB is a good minimum
    return basesize * int(2 ** math.log10(expected_mb))


def _limit_es(expected_mb):
    """Protection against creating too small or too large chunks."""

    if expected_mb < 1:  # < 1 MB
        expected_mb = 1
    elif expected_mb > 10 ** 7:  # > 10 TB
        expected_mb = 10 ** 7
    return expected_mb


def _calc_chunksize(expected_mb):
    """Compute the optimum HDF5 chunksize for I/O purposes.

    Rational: HDF5 takes the data in bunches of chunksize length to write the
    on disk. A BTree in memory is used to map structures on disk. The more
    chunks that are allocated for a dataset the larger the B-tree. Large
    B-trees take memory and causes file storage overhead as well as more disk
    I/O and higher contention for the meta data cache.  You have to balance
    between memory and I/O overhead (small B-trees) and time to access to data
    (big B-trees). The tuning of the chunksize parameter affects the
    performance and the memory consumed. This is based on my own experiments
    and, as always, your mileage may vary.
    """

    expected_mb = _limit_es(expected_mb)
    zone = int(math.log10(expected_mb))
    expected_mb = 10 ** zone
    chunksize = _csformula(expected_mb)
    # XXX: Multiply by 8 seems optimal for sequential access
    return chunksize * 24


def _rowsize(shape, maindim, itemsize):
    """"The size of the rows in bytes in dimensions orthogonal to *maindim*."

    shape:
        Shape of the sample to fit in the row

    maindim:
        The dimension along which iterators work. Its value is 0 (i.e. the first
        dimension) when the dataset is not extendable, and self.extdim (where
        available) for extendable ones.

    itemsize:
        nbytes of each element

    The meaning of *atomic* is that individual elements of a cell can not be
    extracted directly by indexing (i.e.  __getitem__()) the dataset; e.g. if a
    dataset has shape (2, 2) and its atoms have shape (3,), to get the third
    element of the cell at (1, 0) one should use dataset[1,0][2] instead of
    dataset[1,0,2].
    """
    rowsize = itemsize
    for i, dim in enumerate(shape):
        if i != maindim:
            rowsize *= dim
    return rowsize


def calc_chunkshape(shape, expectedrows, itemsize, maindim):
    """Calculate the shape for the HDF5 chunk.

    shape:
        Shape of the sample to fit in the row

    expectedrows:
        how many samples will fit into the file container

    itemsize:
        nbytes of each element

    maindim:
        The dimension along which iterators work. Its value is 0 (i.e. the first
        dimension) when the dataset is not extendable, and self.extdim (where
        available) for extendable ones.

        may want to set to shape.index(max(shape))
    """

    # In case of a scalar shape, return the unit chunksize
    if shape == ():
        return (SizeType(1),)

    MB = 1024 * 1024
    # if shape is sufficiently small, no need to further chunk
    # At time of writing, set to be less than 1MB since that is
    # the limit to hdf5 chunk cache.
    if ((np.prod(shape) * itemsize) < MB) and (shape != ()):
        return shape

    # Compute the chunksize
    rsize = _rowsize(shape, maindim, itemsize)
    expected_mb = (expectedrows * rsize) // MB
    chunksize = _calc_chunksize(expected_mb)

    # Compute the chunknitems
    chunknitems = chunksize // itemsize
    # Safeguard against itemsizes being extremely large
    if chunknitems == 0:
        chunknitems = 1
    chunkshape = list(shape)
    # Check whether trimming the main dimension is enough
    chunkshape[maindim] = 1
    newchunknitems = np.prod(chunkshape, dtype=SizeType)
    if newchunknitems <= chunknitems:
        chunkshape[maindim] = chunknitems // newchunknitems
    else:
        # No, so start trimming other dimensions as well
        for j in range(len(chunkshape)):
            # Check whether trimming this dimension is enough
            chunkshape[j] = 1
            newchunknitems = np.prod(chunkshape, dtype=SizeType)
            if newchunknitems <= chunknitems:
                chunkshape[j] = chunknitems // newchunknitems
                break
        else:
            # Ops, we ran out of the loop without a break
            # Set the last dimension to chunknitems
            chunkshape[-1] = chunknitems

    # safeguard outputing chunks which are larger than shape
    if chunkshape[maindim] > shape[maindim]:
        chunkshape[maindim] = shape[maindim]

    return tuple(SizeType(s) for s in chunkshape)



"""
Hangar's external plugin system is designed to make it flexible for users to
write custom plugins for custom data formats. External plugins should be python
installables and should make itself discoverable using package meta data. A
`detailed documentation <https://packaging.python.org/guides/creating-and-discovering-plugins/#using-package-metadata>`_
can be found in the official python doc. But for a headstart and to avoid going
through this somewhat complex process, we have made a `cookiecutter
<https://github.com/tensorwerk/cookiecutter-hangar-external-plugin>`_ package.
All the hangar plugins follow the naming standard similar to Flask plugins i.e
`hangar_pluginName`
"""

import os


class BasePlugin(object):
    """Base plugin class from where all the external plugins should be inherited.

    Child classes can have four methods to expose - ``load``, ``save``,
    ``show`` and ``board_show``. These are considered as valid methods and
    should be passed as the first argument while initializing the parent from
    child. Child should also inform the parent about the acceptable file
    formats by passing that as second argument. :class:`.BasePlugin` accepts
    ``provides`` and ``accepts`` on init and exposes them which is then used by
    plugin manager while loading the modules. BasePlugin also provides
    ``sample_name`` function to figure out the sample name from the file path.
    This function is used by ``load`` method to return the sample name which is
    then used by hangar as a key to save the data
    """
    def __init__(self, provides, accepts):
        if not provides:
            raise ValueError("Argument ``provides`` cannot be empty")
        if not accepts:
            raise ValueError("Argument ``accepts`` cannot be empty")
        self._provides = provides
        self._accepts = accepts

    @property
    def provides(self):
        return self._provides

    @property
    def accepts(self):
        return self._accepts

    def load(self, fpath, *args, **kwargs):
        """Load some data file on disk to recover it in :class:`numpy.ndarray` form.

        Loads the data provided from the disk for the file path given and
        returns the data as :class:`numpy.ndarray` and name of the data sample.
        Names returned from this function will be used by the import cli system
        as the key for the returned data. This function can return either a
        single :class:`numpy.ndarray`, sample name, combination, or a generator
        that produces one of the the above combinations. This helps when the
        input file is not a single data entry like an image but has multiple
        data points like CSV files.

        An example implementation that returns a single data point:

        .. code-block:: python

            def load(self, fpath, *args, **kwargs):
                data = create_np_array('myimg.jpg')
                name = create_sample_name('myimg.jpg')  # could use `self.sample_name`
                return data, name

        An example implementation that returns a generator could look like this:

        .. code-block:: python

            def load(self, fpath, *args, **kwargs):
                for i, line in enumerate('myfile.csv'):
                    data = create_np_array(line)
                    name = create_sample_name(fpath, i)
                    yield data, name
        """
        raise NotImplementedError

    def save(self, arr, outdir, sample_detail, extension, *args, **kwargs):
        """Save data in a :class:`numpy.ndarray` to a specific file format on disk.

        If the plugin is developed for files like CSV, JSON, etc - where
        multiple data entry would go to the same file - this should check
        whether the file exist already and weather it should modify / append
        the new data entry to the structure, instead of overwriting it or
        throwing an exception.

        Note
        ----
        Name of the file and the whole path to save the data should be constructed
        by this function. This can be done using the information gets as arguments
        such as, ``outdir``, ``sample_detail`` and ``extension``. It has been
        offloaded to this function instead of handling it before because, decisions
        like whether the multiple data entry should go to a single file or mutltpile
        file cannot be predicted before hand as are always data specific (and hence
        plugin specific)

        Note
        ----
        If the call to this function is initiated by the CLI, ``sample_detail`` argument
        will be a string formatted as `sample_name_type:sample_name`. For example, if
        the sample name is `sample1` (and type of sample name is `str`) then
        ``sample_detail`` will be `str:sample1`. This is to avoid the ambiguity that
        could arise by having both integer and string form of numerical as the sample
        name (ex: if column[123] and column["123"] exist). Formatting
        ``sample_detail`` to make a proper filename (not necessary) is upto the
        plugin developer.
        """
        raise NotImplementedError

    def show(self, arr, *args, **kwargs):
        """Show/Display the data to the user.

        This function should process the input :class:`numpy.ndarray` and show
        that to the user using a data dependant display mechanism. A good
        example for such a system is ``matplotlib.pyplot``'s ``plt.show``,
        which displays the image data inline in the running terminal / kernel
        ui.
        """
        raise NotImplementedError

    def board_show(self, arr, *args, **kwargs):
        """Show/display data in hangarboard format.

        Hangarboard is capable of displaying three most common data formats:
        image, text and audio. This function should process the input
        :class:`numpy.ndarray` data and convert it to any of the supported
        formats.
        """
        raise NotImplementedError

    @staticmethod
    def sample_name(fpath: os.PathLike) -> str:
        """Sample the name from file path.

        This function comes handy since the :meth:`.load` method needs to
        ``yield`` or ``return`` both data and sample name. If there no specific
        requirements regarding sample name creation, you can use this function
        which removes the extension from the file name and returns just the
        name. For example, if filepath is ``/path/to/myfile.ext``, then it
        returns ``myfile``

        Parameters
        ----------
        fpath : os.PathLike
            Path to the file which is being loaded by `load`
        """
        return os.path.splitext(os.path.basename(fpath))[0]



from collections import Counter
from typing import MutableMapping

import lmdb


class TxnRegisterSingleton(type):
    _instances = {}
    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(TxnRegisterSingleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class TxnRegister(metaclass=TxnRegisterSingleton):
    """Singleton to manage transaction thread safety in lmdb databases.

    This is essentailly a reference counting transaction register, lots of room
    for improvement here.
    """

    def __init__(self):
        self.WriterAncestors = Counter()
        self.ReaderAncestors = Counter()
        self.WriterTxn: MutableMapping[lmdb.Environment, lmdb.Transaction] = {}
        self.ReaderTxn: MutableMapping[lmdb.Environment, lmdb.Transaction] = {}

    @property
    def _debug_(self):  # pragma: no cover
        return {
            '__class__': self.__class__,
            'WriterAncestors': self.WriterAncestors,
            'ReaderAncestors': self.ReaderAncestors,
            'WriterTxn': self.WriterTxn,
            'ReaderTxn': self.ReaderTxn,
        }

    def begin_writer_txn(self, lmdbenv: lmdb.Environment,
                         buffer: bool = False) -> lmdb.Transaction:
        """Start a write enabled transaction on the given environment

        If multiple write transactions are requested for the same handle, only
        one instance of the transaction handle will be returened, and will not
        close until all operations on that handle have requested to close

        Parameters
        ----------
        lmdbenv : lmdb.Environment
            the environment to open the transaction on
        buffer : bool, optional
            if buffer objects should be used (the default is False, which does
            not use buffers)

        Returns
        -------
        lmdb.Transaction
            transaction handle to perform operations on
        """
        if self.WriterAncestors[lmdbenv] == 0:
            self.WriterTxn[lmdbenv] = lmdbenv.begin(write=True, buffers=buffer)
        self.WriterAncestors[lmdbenv] += 1
        return self.WriterTxn[lmdbenv]

    def begin_reader_txn(self, lmdbenv: lmdb.Environment,
                         buffer: bool = False) -> lmdb.Transaction:
        """Start a reader only txn for the given environment

        If there a read-only transaction for the same environment already exists
        then the same reader txn handle will be returned, and will not close
        until all operations on that handle have said they are finished.

        Parameters
        ----------
        lmdbenv : lmdb.Environment
            the environment to start the transaction in.
        buffer : bool, optional
            weather a buffer transaction should be used (the default is False,
            which means no buffers are returned)

        Returns
        -------
        lmdb.Transaction
            handle to the lmdb transaction.
        """
        if self.ReaderAncestors[lmdbenv] == 0:
            self.ReaderTxn[lmdbenv] = lmdbenv.begin(write=False, buffers=buffer)
        self.ReaderAncestors[lmdbenv] += 1
        return self.ReaderTxn[lmdbenv]

    def commit_writer_txn(self, lmdbenv: lmdb.Environment) -> bool:
        """Commit changes made in a write-enable transaction handle

        As multiple objects can have references to the same open transaction handle,
        the data is not actually committed until all open transactions have called
        the commit method.

        Parameters
        ----------
        lmdbenv : lmdb.Environment
            the environment handle used to open the transaction

        Raises
        ------
        RuntimeError
            If the internal reference counting gets out of sync

        Returns
        -------
        bool
            True if this operation actually committed, otherwise false
            if other objects have references to the same (open) handle
        """
        ancestors = self.WriterAncestors[lmdbenv]
        if ancestors == 0:
            msg = f'hash ancestors are zero but commit called on {lmdbenv}'
            raise RuntimeError(msg)
        elif ancestors == 1:
            self.WriterTxn[lmdbenv].commit()
            self.WriterTxn.__delitem__(lmdbenv)
            ret = True
        else:
            ret = False
        self.WriterAncestors[lmdbenv] -= 1
        return ret

    def abort_reader_txn(self, lmdbenv: lmdb.Environment) -> bool:
        """Request to close a read-only transaction handle

        As multiple objects can have references to the same open transaction
        handle, the transaction is not actuall aborted until all open transactions
        have called the abort method


        Parameters
        ----------
        lmdbenv : lmdb.Environment
            the environment handle used to open the transaction

        Raises
        ------
        RuntimeError
            If the internal reference counting gets out of sync.

        Returns
        -------
        bool
            True if this operation actually aborted the transaction,
            otherwise False if other objects have references to the same (open)
            handle.
        """
        ancestors = self.ReaderAncestors[lmdbenv]
        if ancestors == 0:
            raise RuntimeError(f'hash ancestors are zero but abort called')
        elif ancestors == 1:
            self.ReaderTxn[lmdbenv].abort()
            self.ReaderTxn.__delitem__(lmdbenv)
            ret = True
        else:
            ret = False
        self.ReaderAncestors[lmdbenv] -= 1
        return ret



import atexit
from pathlib import Path
import weakref
from contextlib import suppress, ExitStack
from uuid import uuid4
from typing import Optional, Union

import numpy as np
import lmdb

from .mixins import GetMixin, CheckoutDictIteration
from .columns import (
    ColumnTxn,
    Columns,
    generate_nested_column,
    generate_flat_column,
)
from .diff import ReaderUserDiff, WriterUserDiff
from .merger import select_merge_algorithm
from .records import commiting, hashs, heads, summarize
from .typesystem import (
    NdarrayFixedShape,
    NdarrayVariableShape,
    StringVariableShape,
    BytesVariableShape,
)
from .utils import is_suitable_user_key, is_ascii
from .records import (
    schema_db_key_from_column,
    schema_hash_record_db_val_from_spec,
    schema_hash_db_key_from_digest,
    schema_record_db_val_from_digest,
)


class ReaderCheckout(GetMixin, CheckoutDictIteration):
    """Checkout the repository as it exists at a particular branch.

    This class is instantiated automatically from a repository checkout
    operation. This object will govern all access to data and interaction methods
    the user requests.

        >>> co = repo.checkout()
        >>> isinstance(co, ReaderCheckout)
        True

    If a commit hash is provided, it will take precedent over the branch name
    parameter. If neither a branch not commit is specified, the staging
    environment's base branch ``HEAD`` commit hash will be read.

        >>> co = repo.checkout(commit='foocommit')
        >>> co.commit_hash
        'foocommit'
        >>> co.close()
        >>> co = repo.checkout(branch='testbranch')
        >>> co.commit_hash
        'someothercommithashhere'
        >>> co.close()

    Unlike :class:`WriterCheckout`, any number of :class:`ReaderCheckout`
    objects can exist on the repository independently. Like the
    ``write-enabled`` variant, the :meth:`close` method should be called after
    performing the necessary operations on the repo. However, as there is no
    concept of a ``lock`` for ``read-only`` checkouts, this is just to free up
    memory resources, rather than changing recorded access state.

    In order to reduce the chance that the python interpreter is shut down
    without calling :meth:`close`,  - a common mistake during ipython / jupyter
    sessions - an `atexit <https://docs.python.org/3/library/atexit.html>`_
    hook is registered to :meth:`close`. If properly closed by the user, the
    hook is unregistered after completion with no ill effects. So long as a the
    process is NOT terminated via non-python ``SIGKILL``, fatal internal python
    error, or or special ``os exit`` methods, cleanup will occur on interpreter
    shutdown and resources will be freed. If a non-handled termination method
    does occur, the implications of holding resources varies on a per-OS basis.
    While no risk to data integrity is observed, repeated misuse may require a
    system reboot in order to achieve expected performance characteristics.
    """

    def __init__(self,
                 base_path: Path,
                 dataenv: lmdb.Environment,
                 hashenv: lmdb.Environment,
                 branchenv: lmdb.Environment,
                 refenv: lmdb.Environment,
                 commit: str):
        """Developer documentation of init method.

        Parameters
        ----------
        base_path : Path
            directory path to the Hangar repository on disk
        dataenv : lmdb.Environment
            db where the checkout record data is unpacked and stored.
        hashenv : lmdb.Environment
            db where the hash records are stored.
        branchenv : lmdb.Environment
            db where the branch records are stored.
        refenv : lmdb.Environment
            db where the commit references are stored.
        commit : str
            specific commit hash to checkout
        """
        self._commit_hash = commit
        self._repo_path = base_path
        self._dataenv = dataenv
        self._hashenv = hashenv
        self._branchenv = branchenv
        self._refenv = refenv
        self._enter_count = 0
        self._stack: Optional[ExitStack] = None

        self._columns = Columns._from_commit(
            repo_pth=self._repo_path,
            hashenv=self._hashenv,
            cmtrefenv=self._dataenv)
        self._differ = ReaderUserDiff(
            commit_hash=self._commit_hash,
            branchenv=self._branchenv,
            refenv=self._refenv)
        atexit.register(self.close)

    def _repr_pretty_(self, p, cycle):
        """pretty repr for printing in jupyter notebooks
        """
        self._verify_alive()
        res = f'Hangar {self.__class__.__name__}\
                \n    Writer       : False\
                \n    Commit Hash  : {self._commit_hash}\
                \n    Num Columns  : {len(self)}\n'
        p.text(res)

    def __repr__(self):
        self._verify_alive()
        res = f'{self.__class__}('\
              f'base_path={self._repo_path} '\
              f'dataenv={self._dataenv} '\
              f'hashenv={self._hashenv} '\
              f'commit={self._commit_hash})'
        return res

    def __enter__(self):
        self._verify_alive()
        with ExitStack() as stack:
            if self._enter_count == 0:
                stack.enter_context(self._columns)
            self._enter_count += 1
            self._stack = stack.pop_all()
        return self

    def __exit__(self, *exc):
        self._stack.close()
        self._enter_count -= 1

    def _verify_alive(self):
        """Validates that the checkout object has not been closed

        Raises
        ------
        PermissionError
            if the checkout was previously close
        """
        if not hasattr(self, '_columns'):
            err = PermissionError(
                f'Unable to operate on past checkout objects which have been '
                f'closed. No operation occurred. Please use a new checkout.')
            raise err from None

    @property
    def _is_conman(self) -> bool:
        self._verify_alive()
        return bool(self._enter_count)

    @property
    def columns(self) -> Columns:
        """Provides access to column interaction object.

        Can be used to either return the columns accessor for all elements or
        a single column instance by using dictionary style indexing.

            >>> co = repo.checkout(write=False)
            >>> len(co.columns)
            1
            >>> print(co.columns.keys())
            ['foo']
            >>> fooCol = co.columns['foo']
            >>> fooCol.dtype
            np.fooDtype
            >>> cols = co.columns
            >>> fooCol = cols['foo']
            >>> fooCol.dtype
            np.fooDtype
            >>> fooCol = cols.get('foo')
            >>> fooCol.dtype
            np.fooDtype

        .. seealso::

            The class :class:`~.columns.column.Columns` contains all methods
            accessible by this property accessor

        Returns
        -------
        :class:`~.columns.column.Columns`
            the columns object which behaves exactly like a
            columns accessor class but which can be invalidated when the writer
            lock is released.
        """
        self._verify_alive()
        return self._columns

    @property
    def diff(self) -> ReaderUserDiff:
        """Access the differ methods for a read-only checkout.

        .. seealso::

            The class :class:`ReaderUserDiff` contains all methods accessible
            by this property accessor

        Returns
        -------
        ReaderUserDiff
            weakref proxy to the differ object (and contained methods) which behaves
            exactly like the differ class but which can be invalidated when the
            writer lock is released.
        """
        self._verify_alive()
        wr = weakref.proxy(self._differ)
        return wr

    @property
    def commit_hash(self) -> str:
        """Commit hash this read-only checkout's data is read from.

            >>> co = repo.checkout()
            >>> co.commit_hash
            foohashdigesthere

        Returns
        -------
        str
            commit hash of the checkout
        """
        self._verify_alive()
        return self._commit_hash

    def log(self,
            branch: str = None,
            commit: str = None,
            *,
            return_contents: bool = False,
            show_time: bool = False,
            show_user: bool = False) -> Optional[dict]:
        """Displays a pretty printed commit log graph to the terminal.

        .. note::

            For programatic access, the return_contents value can be set to true
            which will retrieve relevant commit specifications as dictionary
            elements.

        if Neither `branch` nor `commit` arguments are supplied, the commit
        digest of the currently reader checkout will be used as default.

        Parameters
        ----------
        branch : str, optional
            The name of the branch to start the log process from. (Default value
            = None)
        commit : str, optional
            The commit hash to start the log process from. (Default value = None)
        return_contents : bool, optional, kwarg only
            If true, return the commit graph specifications in a dictionary
            suitable for programatic access/evaluation.
        show_time : bool, optional, kwarg only
            If true and return_contents is False, show the time of each commit
            on the printed log graph
        show_user : bool, optional, kwarg only
            If true and return_contents is False, show the committer of each
            commit on the printed log graph
        Returns
        -------
        Optional[dict]
            Dict containing the commit ancestor graph, and all specifications.
        """
        self._verify_alive()
        if (branch is None) and (commit is None):
            commit = self.commit_hash
        res = summarize.log(branchenv=self._branchenv,
                            refenv=self._refenv,
                            branch=branch,
                            commit=commit,
                            return_contents=return_contents,
                            show_time=show_time,
                            show_user=show_user)
        return res

    def close(self) -> None:
        """Gracefully close the reader checkout object.

        Though not strictly required for reader checkouts (as opposed to
        writers), closing the checkout after reading will free file handles and
        system resources, which may improve performance for repositories with
        multiple simultaneous read checkouts.
        """
        self._verify_alive()
        if isinstance(self._stack, ExitStack):
            self._stack.close()

        self._columns._destruct()
        for attr in list(self.__dict__.keys()):
            delattr(self, attr)
        atexit.unregister(self.close)
        return


# --------------- Write enabled checkout ---------------------------------------


class WriterCheckout(GetMixin, CheckoutDictIteration):
    """Checkout the repository at the head of a given branch for writing.

    This is the entry point for all writing operations to the repository, the
    writer class records all interactions in a special ``"staging"`` area,
    which is based off the state of the repository as it existed at the
    ``HEAD`` commit of a branch.

        >>> co = repo.checkout(write=True)
        >>> co.branch_name
        'master'
        >>> co.commit_hash
        'masterheadcommithash'
        >>> co.close()

    At the moment, only one instance of this class can write data to the
    staging area at a time. After the desired operations have been completed,
    it is crucial to call :meth:`close` to release the writer lock. In
    addition, after any changes have been made to the staging area, the branch
    ``HEAD`` cannot be changed. In order to checkout another branch ``HEAD``
    for writing, you must either :meth:`commit` the changes, or perform a
    hard-reset of the staging area to the last commit via
    :meth:`reset_staging_area`.

    In order to reduce the chance that the python interpreter is shut down
    without calling :meth:`close`, which releases the writer lock - a common
    mistake during ipython / jupyter sessions - an `atexit
    <https://docs.python.org/3/library/atexit.html>`_ hook is registered to
    :meth:`close`. If properly closed by the user, the hook is unregistered
    after completion with no ill effects. So long as a the process is NOT
    terminated via non-python SIGKILL, fatal internal python error, or or
    special os exit methods, cleanup will occur on interpreter shutdown and the
    writer lock will be released. If a non-handled termination method does
    occur, the :meth:`~.Repository.force_release_writer_lock` method must be
    called manually when a new python process wishes to open the writer
    checkout.
    """

    def __init__(self,
                 repo_pth: Path,
                 branch_name: str,
                 hashenv: lmdb.Environment,
                 refenv: lmdb.Environment,
                 stageenv: lmdb.Environment,
                 branchenv: lmdb.Environment,
                 stagehashenv: lmdb.Environment,
                 mode: str = 'a'):
        """Developer documentation of init method.

        Parameters
        ----------
        repo_pth : Path
            local file path of the repository.
        branch_name : str
            name of the branch whose ``HEAD`` commit will for the starting state
            of the staging area.
        hashenv lmdb.Environment
            db where the hash records are stored.
        refenv : lmdb.Environment
            db where the commit record data is unpacked and stored.
        stageenv : lmdb.Environment
            db where the stage record data is unpacked and stored.
        branchenv : lmdb.Environment
            db where the head record data is unpacked and stored.
        stagehashenv: lmdb.Environment
            db where the staged hash record data is stored.
        mode : str, optional
            open in write or read only mode, default is 'a' which is write-enabled.
        """
        self._enter_count = 0
        self._repo_path: Path = repo_pth
        self._branch_name = branch_name
        self._writer_lock = str(uuid4())
        self._stack: Optional[ExitStack] = None

        self._refenv = refenv
        self._hashenv = hashenv
        self._stageenv = stageenv
        self._branchenv = branchenv
        self._stagehashenv = stagehashenv

        self._columns: Optional[Columns] = None
        self._differ: Optional[WriterUserDiff] = None
        self._setup()
        atexit.register(self.close)

    def _repr_pretty_(self, p, cycle):
        """pretty repr for printing in jupyter notebooks
        """
        self._verify_alive()
        res = f'Hangar {self.__class__.__name__}\
                \n    Writer       : True\
                \n    Base Branch  : {self._branch_name}\
                \n    Num Columns  : {len(self)}\n'
        p.text(res)

    def __repr__(self):
        self._verify_alive()
        res = f'{self.__class__}('\
              f'base_path={self._repo_path} '\
              f'branch_name={self._branch_name} ' \
              f'hashenv={self._hashenv} '\
              f'refenv={self._refenv} '\
              f'stageenv={self._stageenv} '\
              f'branchenv={self._branchenv})\n'
        return res

    def __enter__(self):
        self._verify_alive()
        with ExitStack() as stack:
            if self._enter_count == 0:
                stack.enter_context(self._columns)
            self._enter_count += 1
            self._stack = stack.pop_all()
        return self

    def __exit__(self, *exc):
        self._stack.close()
        self._enter_count -= 1

    @property
    def _is_conman(self):
        self._verify_alive()
        return bool(self._enter_count)

    def _verify_alive(self):
        """Ensures that this class instance holds the writer lock in the database.

        Raises
        ------
        PermissionError
            If the checkout was previously closed (no :attr:``_writer_lock``)
            or if the writer lock value does not match that recorded in the
            branch db
        """
        if not hasattr(self, '_writer_lock'):
            with suppress(AttributeError):
                self._columns._destruct()
                del self._columns
            with suppress(AttributeError):
                del self._differ
            err = f'Unable to operate on past checkout objects which have been '\
                  f'closed. No operation occurred. Please use a new checkout.'
            raise PermissionError(err) from None

        try:
            heads.acquire_writer_lock(self._branchenv, self._writer_lock)
        except Exception as e:
            with suppress(AttributeError):
                self._columns._destruct()
                del self._columns
            with suppress(AttributeError):
                del self._differ
            raise e from None

    def _setup(self):
        """setup the staging area appropriately for a write enabled checkout.

        On setup, we cannot be sure what branch the staging area was previously
        checked out on, and we cannot be sure if there are any 'uncommitted
        changes' in the staging area (ie. the staging area is ``DIRTY``). The
        setup methods here ensure that we can safety make any changes to the
        staging area without overwriting uncommitted changes, and then perform
        the setup steps to checkout staging area state at that point in time.

        Raises
        ------
        ValueError
            if there are changes previously made in the staging area which were
            based on one branch's ``HEAD``, but a different branch was specified to
            be used for the base of this checkout.
        """
        self._verify_alive()
        current_head = heads.get_staging_branch_head(self._branchenv)
        currentDiff = WriterUserDiff(stageenv=self._stageenv,
                                     refenv=self._refenv,
                                     branchenv=self._branchenv,
                                     branch_name=current_head)
        if currentDiff.status() == 'DIRTY':
            if current_head != self._branch_name:
                e = ValueError(
                    f'Unable to check out branch: {self._branch_name} for writing '
                    f'as the staging area has uncommitted changes on branch: '
                    f'{current_head}. Please commit or stash uncommitted changes '
                    f'before checking out a different branch for writing.')
                self.close()
                raise e
        else:
            if current_head != self._branch_name:
                try:
                    cmt = heads.get_branch_head_commit(
                        branchenv=self._branchenv, branch_name=self._branch_name)
                except ValueError as e:
                    self.close()
                    raise e
                commiting.replace_staging_area_with_commit(
                    refenv=self._refenv, stageenv=self._stageenv, commit_hash=cmt)
                heads.set_staging_branch_head(
                    branchenv=self._branchenv, branch_name=self._branch_name)

        self._columns = Columns._from_staging_area(
            repo_pth=self._repo_path,
            hashenv=self._hashenv,
            stageenv=self._stageenv,
            stagehashenv=self._stagehashenv)
        self._differ = WriterUserDiff(
            stageenv=self._stageenv,
            refenv=self._refenv,
            branchenv=self._branchenv,
            branch_name=self._branch_name)

    @property
    def columns(self) -> Columns:
        """Provides access to column interaction object.

        Can be used to either return the columns accessor for all elements or
        a single column instance by using dictionary style indexing.

            >>> co = repo.checkout(write=True)
            >>> cols = co.columns
            >>> len(cols)
            0
            >>> fooCol = co.add_ndarray_column('foo', shape=(10, 10), dtype=np.uint8)
            >>> len(co.columns)
            1
            >>> len(co)
            1
            >>> list(co.columns.keys())
            ['foo']
            >>> list(co.keys())
            ['foo']
            >>> fooCol = co.columns['foo']
            >>> fooCol.dtype
            np.fooDtype
            >>> fooCol = cols.get('foo')
            >>> fooCol.dtype
            np.fooDtype
            >>> 'foo' in co.columns
            True
            >>> 'bar' in co.columns
            False

        .. seealso::

            The class :class:`~.columns.column.Columns` contains all methods
            accessible by this property accessor

        Returns
        -------
        :class:`~.columns.column.Columns`
            the columns object which behaves exactly like a columns accessor
            class but which can be invalidated when the writer lock is
            released.
        """
        self._verify_alive()
        return self._columns

    @property
    def diff(self) -> WriterUserDiff:
        """Access the differ methods which are aware of any staged changes.

        .. seealso::

            The class :class:`hangar.diff.WriterUserDiff` contains all methods
            accessible by this property accessor

        Returns
        -------
        WriterUserDiff
            weakref proxy to the differ object (and contained methods) which
            behaves exactly like the differ class but which can be invalidated
            when the writer lock is released.
        """
        self._verify_alive()
        wr = weakref.proxy(self._differ)
        return wr

    @property
    def branch_name(self) -> str:
        """Branch this write enabled checkout's staging area was based on.

        Returns
        -------
        str
            name of the branch whose commit ``HEAD`` changes are staged from.
        """
        self._verify_alive()
        return self._branch_name

    @property
    def commit_hash(self) -> str:
        """Commit hash which the staging area of `branch_name` is based on.

        Returns
        -------
        str
            commit hash
        """
        self._verify_alive()
        cmt = heads.get_branch_head_commit(branchenv=self._branchenv,
                                           branch_name=self._branch_name)
        return cmt

    def log(self,
            branch: str = None,
            commit: str = None,
            *,
            return_contents: bool = False,
            show_time: bool = False,
            show_user: bool = False) -> Optional[dict]:
        """Displays a pretty printed commit log graph to the terminal.

        .. note::

            For programatic access, the return_contents value can be set to true
            which will retrieve relevant commit specifications as dictionary
            elements.

        if Neither `branch` nor `commit` arguments are supplied, the branch which
        is currently checked out for writing will be used as default.

        Parameters
        ----------
        branch : str, optional
            The name of the branch to start the log process from. (Default value
            = None)
        commit : str, optional
            The commit hash to start the log process from. (Default value = None)
        return_contents : bool, optional, kwarg only
            If true, return the commit graph specifications in a dictionary
            suitable for programatic access/evaluation.
        show_time : bool, optional, kwarg only
            If true and return_contents is False, show the time of each commit
            on the printed log graph
        show_user : bool, optional, kwarg only
            If true and return_contents is False, show the committer of each
            commit on the printed log graph
        Returns
        -------
        Optional[dict]
            Dict containing the commit ancestor graph, and all specifications.
        """
        self._verify_alive()
        if (branch is None) and (commit is None):
            branch = self.branch_name
        res = summarize.log(branchenv=self._branchenv,
                            refenv=self._refenv,
                            branch=branch,
                            commit=commit,
                            return_contents=return_contents,
                            show_time=show_time,
                            show_user=show_user)
        return res

    def add_str_column(self,
                       name: str,
                       contains_subsamples: bool = False,
                       *,
                       backend: Optional[str] = None,
                       backend_options: Optional[dict] = None):
        """Initializes a :class:`str` container column

        Columns are created in order to store some arbitrary collection of data
        pieces. In this case, we store :class:`str` data. Items need not be
        related to each-other in any direct capacity; the only criteria hangar
        requires is that all pieces of data stored in the column have a
        compatible schema with each-other (more on this below). Each piece of
        data is indexed by some key (either user defined or automatically
        generated depending on the user's preferences). Both single level
        stores (sample keys mapping to data on disk) and nested stores (where
        some sample key maps to an arbitrary number of subsamples, in turn each
        pointing to some piece of store data on disk) are supported.

        All data pieces within a column have the same data type. For
        :class:`str` columns, there is no distinction between
        ``'variable_shape'`` and ``'fixed_shape'`` schema types. Values are
        allowed to take on a value of any size so long as the datatype and
        contents are valid for the schema definition.

        Parameters
        ----------
        name : str
            Name assigned to the column
        contains_subsamples : bool, optional
            True if the column column should store data in a nested structure.
            In this scheme, a sample key is used to index an arbitrary number
            of subsamples which map some (sub)key to a piece of data. If False,
            sample keys map directly to a single piece of data; essentially
            acting as a single level key/value store. By default, False.
        backend : Optional[str], optional
            ADVANCED USERS ONLY, backend format code to use for column data. If
            None, automatically inferred and set based on data shape and type.
            by default None
        backend_options : Optional[dict], optional
            ADVANCED USERS ONLY, filter opts to apply to column data. If None,
            automatically inferred and set based on data shape and type.
            by default None

        Returns
        -------
        :class:`~.columns.column.Columns`
            instance object of the initialized column.
        """
        self._verify_alive()
        if self.columns._any_is_conman() or self._is_conman:
            raise PermissionError('Not allowed while context manager is used.')

        # ------------- Checks for argument validity --------------------------

        try:
            if (not is_suitable_user_key(name)) or (not is_ascii(name)):
                raise ValueError(
                    f'Column name provided: `{name}` is invalid. Can only contain '
                    f'alpha-numeric or "." "_" "-" ascii characters (no whitespace). '
                    f'Must be <= 64 characters long')
            if name in self._columns:
                raise LookupError(f'Column already exists with name: {name}.')
            if not isinstance(contains_subsamples, bool):
                raise ValueError(f'contains_subsamples argument must be bool, '
                                 f'not type {type(contains_subsamples)}')
        except (ValueError, LookupError) as e:
            raise e from None

        # ---------- schema validation handled automatically by typesystem ----

        layout = 'nested' if contains_subsamples else 'flat'
        schema = StringVariableShape(
            dtype=str, column_layout=layout, backend=backend, backend_options=backend_options)

        # ------------------ create / return new column -----------------------

        col = self._initialize_new_column(
            column_name=name, column_layout=layout, schema=schema)
        return col

    def add_bytes_column(self,
                         name: str,
                         contains_subsamples: bool = False,
                         *,
                         backend: Optional[str] = None,
                         backend_options: Optional[dict] = None):
        """Initializes a :class:`bytes` container column

        Columns are created in order to store some arbitrary collection of data
        pieces. In this case, we store :class:`bbytes` data. Items need not be
        related to each-other in any direct capacity; the only criteria hangar
        requires is that all pieces of data stored in the column have a
        compatible schema with each-other (more on this below). Each piece of
        data is indexed by some key (either user defined or automatically
        generated depending on the user's preferences). Both single level
        stores (sample keys mapping to data on disk) and nested stores (where
        some sample key maps to an arbitrary number of subsamples, in turn each
        pointing to some piece of store data on disk) are supported.

        All data pieces within a column have the same data type. For
        :class:`bytes` columns, there is no distinction between
        ``'variable_shape'`` and ``'fixed_shape'`` schema types. Values are
        allowed to take on a value of any size so long as the datatype and
        contents are valid for the schema definition.

        Parameters
        ----------
        name : str
            Name assigned to the column
        contains_subsamples : bool, optional
            True if the column column should store data in a nested structure.
            In this scheme, a sample key is used to index an arbitrary number
            of subsamples which map some (sub)key to a piece of data. If False,
            sample keys map directly to a single piece of data; essentially
            acting as a single level key/value store. By default, False.
        backend : Optional[str], optional
            ADVANCED USERS ONLY, backend format code to use for column data. If
            None, automatically inferred and set based on data shape and type.
            by default None
        backend_options : Optional[dict], optional
            ADVANCED USERS ONLY, filter opts to apply to column data. If None,
            automatically inferred and set based on data shape and type.
            by default None

        Returns
        -------
        :class:`~.columns.column.Columns`
            instance object of the initialized column.
        """
        self._verify_alive()
        if self.columns._any_is_conman() or self._is_conman:
            raise PermissionError('Not allowed while context manager is used.')

        # ------------- Checks for argument validity --------------------------

        try:
            if (not is_suitable_user_key(name)) or (not is_ascii(name)):
                raise ValueError(
                    f'Column name provided: `{name}` is invalid. Can only contain '
                    f'alpha-numeric or "." "_" "-" ascii characters (no whitespace). '
                    f'Must be <= 64 characters long')
            if name in self._columns:
                raise LookupError(f'Column already exists with name: {name}.')
            if not isinstance(contains_subsamples, bool):
                raise ValueError(f'contains_subsamples argument must be bool, '
                                 f'not type {type(contains_subsamples)}')
        except (ValueError, LookupError) as e:
            raise e from None

        # ---------- schema validation handled automatically by typesystem ----

        layout = 'nested' if contains_subsamples else 'flat'
        schema = BytesVariableShape(
            dtype=bytes, column_layout=layout, backend=backend, backend_options=backend_options)

        # ------------------ create / return new column -----------------------

        col = self._initialize_new_column(
            column_name=name, column_layout=layout, schema=schema)
        return col

    def add_ndarray_column(self,
                           name: str,
                           shape: Optional[Union[int, tuple]] = None,
                           dtype: Optional[np.dtype] = None,
                           prototype: Optional[np.ndarray] = None,
                           variable_shape: bool = False,
                           contains_subsamples: bool = False,
                           *,
                           backend: Optional[str] = None,
                           backend_options: Optional[dict] = None):
        """Initializes a :class:`numpy.ndarray` container column.

        Columns are created in order to store some arbitrary collection of data
        pieces. In this case, we store :class:`numpy.ndarray` data. Items need
        not be related to each-other in any direct capacity; the only criteria
        hangar requires is that all pieces of data stored in the column have a
        compatible schema with each-other (more on this below). Each piece of
        data is indexed by some key (either user defined or automatically
        generated depending on the user's preferences). Both single level
        stores (sample keys mapping to data on disk) and nested stores (where
        some sample key maps to an arbitrary number of subsamples, in turn each
        pointing to some piece of store data on disk) are supported.

        All data pieces within a column have the same data type and number of
        dimensions. The size of each dimension can be either fixed (the default
        behavior) or variable per sample. For fixed dimension sizes, all data
        pieces written to the column must have the same shape & size which was
        specified at the time the column column was initialized. Alternatively,
        variable sized columns can write data pieces with dimensions of any
        size (up to a specified maximum).

        Parameters
        ----------
        name : str
            The name assigned to this column.
        shape : Optional[Union[int, Tuple[int]]]
            The shape of the data samples which will be written in this column.
            This argument and the `dtype` argument are required if a `prototype`
            is not provided, defaults to None.
        dtype : Optional[:class:`numpy.dtype`]
            The datatype of this column. This argument and the `shape` argument
            are required if a `prototype` is not provided., defaults to None.
        prototype : Optional[:class:`numpy.ndarray`]
            A sample array of correct datatype and shape which will be used to
            initialize the column storage mechanisms. If this is provided, the
            `shape` and `dtype` arguments must not be set, defaults to None.
        variable_shape : bool, optional
            If this is a variable sized column. If true, a the maximum shape is
            set from the provided ``shape`` or ``prototype`` argument. Any sample
            added to the column can then have dimension sizes <= to this
            initial specification (so long as they have the same rank as what
            was specified) defaults to False.
        contains_subsamples : bool, optional
            True if the column column should store data in a nested structure.
            In this scheme, a sample key is used to index an arbitrary number of
            subsamples which map some (sub)key to some piece of data. If False,
            sample keys map directly to a single piece of data; essentially
            acting as a single level key/value store. By default, False.
        backend : Optional[str], optional
            ADVANCED USERS ONLY, backend format code to use for column data. If
            None, automatically inferred and set based on data shape and type.
            by default None
        backend_options : Optional[dict], optional
            ADVANCED USERS ONLY, filter opts to apply to column data. If None,
            automatically inferred and set based on data shape and type.
            by default None

        Returns
        -------
        :class:`~.columns.column.Columns`
            instance object of the initialized column.
        """
        self._verify_alive()
        if self.columns._any_is_conman() or self._is_conman:
            raise PermissionError('Not allowed while context manager is used.')

        # ------------- Checks for argument validity --------------------------

        try:
            if (not is_suitable_user_key(name)) or (not is_ascii(name)):
                raise ValueError(
                    f'Column name provided: `{name}` is invalid. Can only contain '
                    f'alpha-numeric or "." "_" "-" ascii characters (no whitespace). '
                    f'Must be <= 64 characters long')
            if name in self.columns:
                raise LookupError(f'Column already exists with name: {name}.')
            if not isinstance(contains_subsamples, bool):
                raise ValueError(f'contains_subsamples is not bool type')

            # If shape/dtype is passed instead of a prototype arg, we use those values
            # to initialize a numpy array prototype. Using a :class:`numpy.ndarray`
            # for specification of dtype / shape params lets us offload much of the
            # required type checking / sanitization of userspace input to libnumpy,
            # rather than attempting to cover all possible cases here.
            if prototype is not None:
                if (shape is not None) or (dtype is not None):
                    raise ValueError(f'cannot set both prototype and shape/dtype args.')
            else:
                prototype = np.zeros(shape, dtype=dtype)
            dtype = prototype.dtype
            shape = prototype.shape
            if not all([x > 0 for x in shape]):
                raise ValueError(f'all dimensions must be sized greater than zero')
        except (ValueError, LookupError) as e:
            raise e from None

        # ---------- schema validation handled automatically by typesystem ----

        column_layout = 'nested' if contains_subsamples else 'flat'
        if variable_shape:
            schema = NdarrayVariableShape(dtype=dtype, shape=shape, column_layout=column_layout,
                                          backend=backend, backend_options=backend_options)
        else:
            schema = NdarrayFixedShape(dtype=dtype, shape=shape, column_layout=column_layout,
                                       backend=backend, backend_options=backend_options)

        # ------------------ create / return new column -----------------------

        col = self._initialize_new_column(
            column_name=name, column_layout=column_layout, schema=schema)
        return col

    def _initialize_new_column(self,
                               column_name: str,
                               column_layout: str,
                               schema) -> Columns:
        """Initialize a column and write spec to record db.

        Parameters
        ----------
        column_name: str
            name of the column
        column_layout: str
            One of ['flat', 'nested'] indicating column layout class to use
            during generation.
        schema: ColumnBase
            schema class instance providing column data spec, schema/column digest,
            data validator / hashing methods, and backend ID / options; all of which
            are needed to successfully create & save the column instance

        Returns
        -------
        Columns
            initialized column class instance.
        """
        # -------- set vals in lmdb only after schema is sure to exist --------

        schema_digest = schema.schema_hash_digest()
        columnSchemaKey = schema_db_key_from_column(column_name, layout=column_layout)
        columnSchemaVal = schema_record_db_val_from_digest(schema_digest)
        hashSchemaKey = schema_hash_db_key_from_digest(schema_digest)
        hashSchemaVal = schema_hash_record_db_val_from_spec(schema.schema)

        txnctx = ColumnTxn(self._stageenv, self._hashenv, self._stagehashenv)
        with txnctx.write() as ctx:
            ctx.dataTxn.put(columnSchemaKey, columnSchemaVal)
            ctx.hashTxn.put(hashSchemaKey, hashSchemaVal, overwrite=False)

        # ------------- create column instance and return to user -------------

        if column_layout == 'nested':
            setup_args = generate_nested_column(
                txnctx=txnctx, column_name=column_name,
                path=self._repo_path, schema=schema, mode='a')
        else:
            setup_args = generate_flat_column(
                txnctx=txnctx, column_name=column_name,
                path=self._repo_path, schema=schema, mode='a')

        self.columns._columns[column_name] = setup_args
        return self.columns[column_name]

    def merge(self, message: str, dev_branch: str) -> str:
        """Merge the currently checked out commit with the provided branch name.

        If a fast-forward merge is possible, it will be performed, and the
        commit message argument to this function will be ignored.

        Parameters
        ----------
        message : str
            commit message to attach to a three-way merge
        dev_branch : str
            name of the branch which should be merge into this branch
            (ie `master`)

        Returns
        -------
        str
            commit hash of the new commit for the `master` branch this checkout
            was started from.
        """
        self._verify_alive()
        commit_hash = select_merge_algorithm(
            message=message,
            branchenv=self._branchenv,
            stageenv=self._stageenv,
            refenv=self._refenv,
            stagehashenv=self._stagehashenv,
            master_branch=self._branch_name,
            dev_branch=dev_branch,
            repo_path=self._repo_path,
            writer_uuid=self._writer_lock)

        for asetHandle in self._columns.values():
            with suppress(KeyError):
                asetHandle._close()

        self._columns = Columns._from_staging_area(
            repo_pth=self._repo_path,
            hashenv=self._hashenv,
            stageenv=self._stageenv,
            stagehashenv=self._stagehashenv)
        self._differ = WriterUserDiff(
            stageenv=self._stageenv,
            refenv=self._refenv,
            branchenv=self._branchenv,
            branch_name=self._branch_name)

        return commit_hash

    def commit(self, commit_message: str) -> str:
        """Commit the changes made in the staging area on the checkout branch.

        Parameters
        ----------
        commit_message : str, optional
            user proved message for a log of what was changed in this commit.
            Should a fast forward commit be possible, this will NOT be added to
            fast-forward ``HEAD``.

        Returns
        -------
        str
            The commit hash of the new commit.

        Raises
        ------
        RuntimeError
            If no changes have been made in the staging area, no commit occurs.
        """
        self._verify_alive()

        open_columns = []
        for column in self._columns.values():
            if column._is_conman:
                open_columns.append(column.column)

        try:
            for column_name in open_columns:
                self._columns[column_name].__exit__()

            if self._differ.status() == 'CLEAN':
                e = RuntimeError('No changes made in staging area. Cannot commit.')
                raise e from None

            self._columns._close()
            commit_hash = commiting.commit_records(message=commit_message,
                                                   branchenv=self._branchenv,
                                                   stageenv=self._stageenv,
                                                   refenv=self._refenv,
                                                   repo_path=self._repo_path)
            # purge recs then reopen file handles so that we don't have to invalidate
            # previous weakproxy references like if we just called :meth:``_setup```
            hashs.clear_stage_hash_records(self._stagehashenv)
            self._columns._open()

        finally:
            for column_name in open_columns:
                self._columns[column_name].__enter__()

        return commit_hash

    def reset_staging_area(self) -> str:
        """Perform a hard reset of the staging area to the last commit head.

        After this operation completes, the writer checkout will automatically
        close in the typical fashion (any held references to :attr:``column``
        or :attr:``metadata`` objects will finalize and destruct as normal), In
        order to perform any further operation, a new checkout needs to be
        opened.

        .. warning::

            This operation is IRREVERSIBLE. all records and data which are note
            stored in a previous commit will be permanently deleted.

        Returns
        -------
        str
            Commit hash of the head which the staging area is reset to.

        Raises
        ------
        RuntimeError
            If no changes have been made to the staging area, No-Op.
        """
        self._verify_alive()
        print(f'Hard reset requested with writer_lock: {self._writer_lock}')

        if self._differ.status() == 'CLEAN':
            e = RuntimeError(f'No changes made in staging area. No reset necessary.')
            raise e from None

        if isinstance(self._stack, ExitStack):
            self._stack.close()
        if hasattr(self._columns, '_destruct'):
            self._columns._destruct()

        hashs.remove_stage_hash_records_from_hashenv(self._hashenv, self._stagehashenv)
        hashs.clear_stage_hash_records(self._stagehashenv)
        hashs.backends_remove_in_process_data(self._repo_path)

        branch_head = heads.get_staging_branch_head(self._branchenv)
        head_commit = heads.get_branch_head_commit(self._branchenv, branch_head)
        if head_commit == '':
            with suppress(ValueError):
                commiting.replace_staging_area_with_commit(refenv=self._refenv,
                                                           stageenv=self._stageenv,
                                                           commit_hash=head_commit)
        else:
            commiting.replace_staging_area_with_commit(refenv=self._refenv,
                                                       stageenv=self._stageenv,
                                                       commit_hash=head_commit)

        self._columns = Columns._from_staging_area(
            repo_pth=self._repo_path,
            hashenv=self._hashenv,
            stageenv=self._stageenv,
            stagehashenv=self._stagehashenv)
        self._differ = WriterUserDiff(
            stageenv=self._stageenv,
            refenv=self._refenv,
            branchenv=self._branchenv,
            branch_name=self._branch_name)
        return head_commit

    def close(self) -> None:
        """Close all handles to the writer checkout and release the writer lock.

        Failure to call this method after the writer checkout has been used
        will result in a lock being placed on the repository which will not
        allow any writes until it has been manually cleared.
        """
        with suppress(lmdb.Error):
            self._verify_alive()

        if isinstance(self._stack, ExitStack):
            self._stack.close()

        if hasattr(self, '_columns'):
            if hasattr(self._columns, '_destruct'):
                self._columns._destruct()

        with suppress(lmdb.Error):
            heads.release_writer_lock(self._branchenv, self._writer_lock)

        for attr in list(self.__dict__.keys()):
            delattr(self, attr)
        atexit.unregister(self.close)
        return
